{
  (* a quick LaTeX pretty-printer for programs *)

  let mots_cles = [ "begin"; "end"; "while"; "do"; "done";
		    "invariant"; "variant"; "assert"; "label"; "as";
		    "if"; "then"; "else"; "for";
		    "fun"; "array"; "of"; "let"; "in";
		    "and"; "or"; "not"; "rec"; "ref";
		    "comment"; "integer"; "goto" (* for find *)
                  ]

  let est_mot_cle s = List.mem s mots_cles

  let debut_ligne = ref true

  let r_underscore = Str.regexp "_"
  let latex_string s = Str.global_replace r_underscore "\\_" s

  let print_mot s = 
    Printf.printf "\\mathit{%s}" (latex_string s)

  let numeroter = ref false
  let numero_ligne () = if !numeroter then print_string "\\PL & "

  let math_mode = ref false

  let r_for_all = Str.regexp "(\\(.*\\):\\(.*\\))"
  let print_for_all s =
    if !math_mode & Str.string_match r_for_all s 0 then
      let v = latex_string (Str.matched_group 1 s) in
      let t = latex_string (Str.matched_group 2 s) in
      Printf.printf "\\forall %s:%s. " v t
    else
      print_string (latex_string s)

  let initial s =
    let n = String.length s in
    Printf.printf "%s " (String.make n '~')
}

let espaces = [ ' ' '\t' ]+

let mot = [ 'a'-'z' 'A'-'Z' '_' ] [ 'a'-'z' 'A'-'Z' '_' '0'-'9' ]*

rule pp = parse
    espaces { if !debut_ligne then begin
		debut_ligne := false; numero_ligne();
		initial (Lexing.lexeme lexbuf)
	      end else
		Printf.printf " ~ ";
	      pp lexbuf
	    }
  | "as" espaces mot { pp lexbuf }
  | mot
    { let mot = Lexing.lexeme lexbuf in
      debut_ligne := false;
      if est_mot_cle mot then 
	Printf.printf "\\kw{%s}" mot
      else
	print_mot mot;
      pp lexbuf
    }
  | mot espaces ":=" espaces { print_mot (Lexing.lexeme lexbuf); 
			       pp lexbuf }
  | '(' mot (',' mot)* ':' mot ')' 
       { print_for_all (Lexing.lexeme lexbuf); pp lexbuf }
  | espaces "<=" espaces { print_string " \\le "; pp lexbuf }
  | espaces "<=" espaces '!' { print_string " \\le\ !"; pp lexbuf }
  | espaces ">=" espaces { print_string " \\ge "; pp lexbuf }
  | espaces ">=" espaces '!' { print_string " \\ge\ !"; pp lexbuf }
  | espaces ('<' | '>') espaces { print_string (Lexing.lexeme lexbuf); 
				  pp lexbuf }
  | espaces '<' espaces '!' { print_string " <\ !";  pp lexbuf } 
  | espaces '>' espaces '!' { print_string " >\ !";  pp lexbuf }
  | '\n' { debut_ligne := true; Printf.printf "\\\\\n"; pp lexbuf }
  | '!' { Printf.printf "\\,!"; pp lexbuf }
  | '@' ("0" | mot) { Printf.printf "_{%s}" (Lexing.lexeme lexbuf); pp lexbuf }
  | '@' { print_string "_@"; pp lexbuf }
  | '{' { math_mode := true; print_string "\\{ ~ "; pp lexbuf }
  | '}' { math_mode := false; print_string "~ \\}"; pp lexbuf }
  | "/\\" { print_string "\\land "; pp lexbuf }
  | "\\/" { print_string "\\lor "; pp lexbuf }
  | "*" { print_string "\\times "; pp lexbuf }
  | "->" { print_string 
	     (if !math_mode then "\\Rightarrow " else "\\rightarrow ");
	   pp lexbuf } 
  | "[!" { Printf.printf "[!"; pp lexbuf }
  | '`' | '#'  { pp lexbuf }
  | _ { print_string (Lexing.lexeme lexbuf); pp lexbuf }
  | eof { () }
{

  let main () =
    if Array.length Sys.argv > 1 & Sys.argv.(1) = "-n" then numeroter := true;
    pp (Lexing.from_channel stdin)

  let _ = Printexc.print main ()

}
